# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

try:
    from beartype.roar import BeartypeCallHintParamViolation

    ValidationErrors = (TypeError, BeartypeCallHintParamViolation)
except ImportError:
    ValidationErrors = TypeError

from typing import (
    Callable,
    Type,
)

from jax import config
import jax.numpy as jnp
import jax.random as jr
import pytest
import tensorflow_probability.substrates.jax.distributions as tfd

# from gpjax.dataset import Dataset
from gpjax.dataset import Dataset
from gpjax.distributions import GaussianDistribution
from gpjax.gps import (
    AbstractPosterior,
    AbstractPrior,
    ConjugatePosterior,
    NonConjugatePosterior,
    Prior,
    construct_posterior,
)
from gpjax.kernels import (
    RBF,
    AbstractKernel,
    Matern52,
)
from gpjax.likelihoods import (
    AbstractLikelihood,
    Bernoulli,
    Gaussian,
    Poisson,
)
from gpjax.mean_functions import (
    AbstractMeanFunction,
    Constant,
    Zero,
)

# Enable Float64 for more stable matrix inversions.
config.update("jax_enable_x64", True)


def test_abstract_prior():
    # Abstract prior should not be able to be instantiated.
    with pytest.raises(TypeError):
        AbstractPrior()


def test_abstract_posterior():
    # Abstract posterior should not be able to be instantiated.
    with pytest.raises(TypeError):
        AbstractPosterior()


@pytest.mark.parametrize("num_datapoints", [1, 10])
@pytest.mark.parametrize("kernel", [RBF, Matern52])
@pytest.mark.parametrize("mean_function", [Zero, Constant])
def test_prior(
    num_datapoints: int,
    kernel: type[AbstractKernel],
    mean_function: Type[AbstractMeanFunction],
) -> None:
    # Create prior.
    prior = Prior(mean_function=mean_function(), kernel=kernel())

    # Check types.
    assert isinstance(prior, Prior)
    assert isinstance(prior, AbstractPrior)

    # Query a marginal distribution at some inputs.
    inputs = jnp.linspace(-3.0, 3.0, num_datapoints).reshape(-1, 1)
    marginal_distribution = prior(inputs)

    # Ensure that the marginal distribution is a Gaussian.
    assert isinstance(marginal_distribution, GaussianDistribution)
    assert isinstance(marginal_distribution, tfd.Distribution)

    # Ensure that the marginal distribution has the correct shape.
    mu = marginal_distribution.mean()
    sigma = marginal_distribution.covariance()
    assert mu.shape == (num_datapoints,)
    assert sigma.shape == (num_datapoints, num_datapoints)


@pytest.mark.parametrize("num_datapoints", [1, 10])
@pytest.mark.parametrize("kernel", [RBF, Matern52])
@pytest.mark.parametrize("mean_function", [Zero, Constant])
def test_conjugate_posterior(
    num_datapoints: int,
    kernel: type[AbstractKernel],
    mean_function: type[AbstractMeanFunction],
) -> None:
    # Create a dataset.
    key = jr.key(123)
    x = jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(num_datapoints, 1))
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define prior.
    prior = Prior(mean_function=mean_function(), kernel=kernel())

    # Define a likelihood.
    likelihood = Gaussian(num_datapoints=num_datapoints)

    # Construct the posterior via the class.
    posterior = ConjugatePosterior(prior=prior, likelihood=likelihood)

    # Check types.
    assert isinstance(posterior, ConjugatePosterior)

    # Query a marginal distribution of the posterior at some inputs.
    inputs = jnp.linspace(-3.0, 3.0, num_datapoints).reshape(-1, 1)
    marginal_distribution = posterior(inputs, D)

    # Ensure that the marginal distribution is a Gaussian.
    assert isinstance(marginal_distribution, GaussianDistribution)
    assert isinstance(marginal_distribution, tfd.Distribution)

    # Ensure that the marginal distribution has the correct shape.
    mu = marginal_distribution.mean()
    sigma = marginal_distribution.covariance()
    assert mu.shape == (num_datapoints,)
    assert sigma.shape == (num_datapoints, num_datapoints)


@pytest.mark.parametrize("num_datapoints", [1, 10])
@pytest.mark.parametrize("kernel", [RBF, Matern52])
@pytest.mark.parametrize("mean_function", [Zero, Constant])
def test_nonconjugate_posterior(
    num_datapoints: int,
    kernel: type[AbstractKernel],
    mean_function: type[AbstractMeanFunction],
) -> None:
    # Create a dataset.
    key = jr.key(123)
    x = jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(num_datapoints, 1))
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define prior.
    prior = Prior(mean_function=mean_function(), kernel=kernel())

    # Define a likelihood.
    likelihood = Bernoulli(num_datapoints=num_datapoints)

    # Construct the posterior via the class.
    posterior = NonConjugatePosterior(prior=prior, likelihood=likelihood)

    # Check types.
    assert isinstance(posterior, NonConjugatePosterior)

    # Check latent values.
    latent_values = jr.normal(posterior.key.value, (num_datapoints, 1))
    assert (posterior.latent.value == latent_values).all()

    # Query a marginal distribution of the posterior at some inputs.
    inputs = jnp.linspace(-3.0, 3.0, num_datapoints).reshape(-1, 1)
    marginal_distribution = posterior(inputs, D)

    # Ensure that the marginal distribution is a Gaussian.
    assert isinstance(marginal_distribution, GaussianDistribution)
    assert isinstance(marginal_distribution, tfd.Distribution)

    # Ensure that the marginal distribution has the correct shape.
    mu = marginal_distribution.mean()
    sigma = marginal_distribution.covariance()
    assert mu.shape == (num_datapoints,)
    assert sigma.shape == (num_datapoints, num_datapoints)


@pytest.mark.parametrize("likelihood", [Bernoulli, Gaussian])
@pytest.mark.parametrize("num_datapoints", [1, 10])
@pytest.mark.parametrize("kernel", [RBF, Matern52])
@pytest.mark.parametrize("mean_function", [Zero, Constant])
def test_posterior_construct(
    likelihood: Type[AbstractLikelihood],
    num_datapoints: int,
    kernel: type[AbstractKernel],
    mean_function: type[AbstractMeanFunction],
) -> None:
    # Define prior.
    prior = Prior(mean_function=mean_function(), kernel=kernel())

    # Construct the posterior via the three methods.
    likelihood: AbstractLikelihood = likelihood(num_datapoints=num_datapoints)
    posterior_mul = prior * likelihood
    posterior_rmul = likelihood * prior
    posterior_manual = construct_posterior(prior=prior, likelihood=likelihood)

    # Ensure that the posterior is the same type in all three cases.
    assert type(posterior_mul) is type(posterior_rmul)
    assert type(posterior_mul) is type(posterior_manual)

    # Ensure we have the correct likelihood and prior.
    assert posterior_mul.likelihood == likelihood
    assert posterior_mul.prior == prior

    # If the likelihood is Gaussian, then the posterior should be conjugate.
    if isinstance(likelihood, Gaussian):
        assert isinstance(posterior_mul, ConjugatePosterior)

    # If the likelihood is Bernoulli or Poisson, then the posterior should be non-conjugate.
    if isinstance(likelihood, (Bernoulli, Poisson)):
        assert isinstance(posterior_mul, NonConjugatePosterior)


@pytest.mark.parametrize("num_datapoints", [1, 5])
@pytest.mark.parametrize("kernel", [RBF, Matern52])
@pytest.mark.parametrize("mean_function", [Zero, Constant])
def test_prior_sample_approx(num_datapoints, kernel, mean_function):
    kern = kernel(n_dims=2, lengthscale=jnp.array([5.0, 1.0]), variance=0.1)
    p = Prior(kernel=kern, mean_function=mean_function())
    key = jr.PRNGKey(123)

    with pytest.raises(ValueError):
        p.sample_approx(-1, key)
    with pytest.raises(ValueError):
        p.sample_approx(0, key)
    with pytest.raises(ValidationErrors):
        p.sample_approx(0.5, key)
    with pytest.raises(ValueError):
        p.sample_approx(1, key, -10)
    with pytest.raises(ValueError):
        p.sample_approx(1, key, 0)
    with pytest.raises(ValidationErrors):
        p.sample_approx(1, key, 0.5)

    sampled_fn = p.sample_approx(1, key, 100)
    assert isinstance(sampled_fn, Callable)  # check type

    x = jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(num_datapoints, 2))
    evals = sampled_fn(x)
    assert evals.shape == (num_datapoints, 1.0)  # check shape

    sampled_fn_2 = p.sample_approx(1, key, 100)
    evals_2 = sampled_fn_2(x)
    max_delta = jnp.max(jnp.abs(evals - evals_2))
    assert max_delta == 0.0  # samples same for same seed

    new_key = jr.key(12345)
    sampled_fn_3 = p.sample_approx(1, new_key, 100)
    evals_3 = sampled_fn_3(x)
    max_delta = jnp.max(jnp.abs(evals - evals_3))
    assert max_delta > 0.01  # samples different for different seed

    # Check validty of samples using Monte-Carlo
    sampled_fn = p.sample_approx(10_000, key, 100)
    sampled_evals = sampled_fn(x)
    approx_mean = jnp.mean(sampled_evals, -1)
    approx_var = jnp.var(sampled_evals, -1)
    true_predictive = p(x)
    true_mean = true_predictive.mean()
    true_var = jnp.diagonal(true_predictive.covariance())
    max_error_in_mean = jnp.max(jnp.abs(approx_mean - true_mean))
    max_error_in_var = jnp.max(jnp.abs(approx_var - true_var))
    assert max_error_in_mean < 0.02  # check that samples are correct
    assert max_error_in_var < 0.05  # check that samples are correct


@pytest.mark.parametrize("num_datapoints", [1, 5])
@pytest.mark.parametrize("kernel", [RBF, Matern52])
@pytest.mark.parametrize("mean_function", [Zero, Constant])
def test_conjugate_posterior_sample_approx(num_datapoints, kernel, mean_function):
    kern = kernel(lengthscale=jnp.array([5.0, 1.0]), variance=0.1)
    p = Prior(kernel=kern, mean_function=mean_function()) * Gaussian(
        num_datapoints=num_datapoints
    )
    key = jr.key(123)

    x = jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(num_datapoints, 2))
    y = (
        jnp.mean(jnp.sin(x), 1, keepdims=True)
        + jr.normal(key=key, shape=(num_datapoints, 1)) * 0.1
    )
    D = Dataset(X=x, y=y)

    # with pytest.raises(ValueError):
    #     p.sample_approx(-1, D, key)
    # with pytest.raises(ValueError):
    #     p.sample_approx(0, D, key)
    # with pytest.raises(ValidationErrors):
    #     p.sample_approx(0.5, D, key)
    # with pytest.raises(ValueError):
    #     p.sample_approx(1, D, key, -10)
    # with pytest.raises(ValueError):
    #     p.sample_approx(1, D, key, 0)
    # with pytest.raises(ValidationErrors):
    #     p.sample_approx(1, D, key, 0.5)

    sampled_fn = p.sample_approx(1, D, key, 100)
    assert isinstance(sampled_fn, Callable)  # check type

    x = jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(num_datapoints, 2))
    evals = sampled_fn(x)
    assert evals.shape == (num_datapoints, 1.0)  # check shape

    sampled_fn_2 = p.sample_approx(1, D, key, 100)
    evals_2 = sampled_fn_2(x)
    max_delta = jnp.max(jnp.abs(evals - evals_2))
    assert max_delta == 0.0  # samples same for same seed

    new_key = jr.key(12345)
    sampled_fn_3 = p.sample_approx(1, D, new_key, 100)
    evals_3 = sampled_fn_3(x)
    max_delta = jnp.max(jnp.abs(evals - evals_3))
    assert max_delta > 0.01  # samples different for different seed

    # Check validty of samples using Monte-Carlo
    sampled_fn = p.sample_approx(10_000, D, key, 100)
    sampled_evals = sampled_fn(x)
    approx_mean = jnp.mean(sampled_evals, -1)
    approx_var = jnp.var(sampled_evals, -1)
    true_predictive = p(x, train_data=D)
    true_mean = true_predictive.mean()
    true_var = jnp.diagonal(true_predictive.covariance())
    max_error_in_mean = jnp.max(jnp.abs(approx_mean - true_mean))
    max_error_in_var = jnp.max(jnp.abs(approx_var - true_var))
    assert max_error_in_mean < 0.02  # check that samples are correct
    assert max_error_in_var < 0.05  # check that samples are correct


if __name__ == "__main__":
    test_conjugate_posterior_sample_approx(10, RBF, Zero)



# Copyright 2024 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from dataclasses import dataclass

from beartype.typing import Mapping
from jaxtyping import Num
import tensorflow_probability.substrates.jax as tfp

from gpjax.dataset import Dataset
from gpjax.decision_making.utility_functions.base import (
    AbstractSinglePointUtilityFunctionBuilder,
    SinglePointUtilityFunction,
)
from gpjax.decision_making.utils import (
    OBJECTIVE,
    get_best_latent_observation_val,
)
from gpjax.gps import ConjugatePosterior
from gpjax.typing import (
    Array,
    KeyArray,
)


@dataclass
class ProbabilityOfImprovement(AbstractSinglePointUtilityFunctionBuilder):
    r"""
    An acquisition function which returns the probability of improvement
    of the objective function over the best observed value.

    More precisely, given a predictive posterior distribution of the objective
    function $`f`$, the probability of improvement at a test point $`x`$ is defined as:
    $$`\text{PI}(x) = \text{Prob}[f(x) < f(x_{\text{best}})]`$$
    where $`x_{\text{best}}`$ is the minimiser of the posterior mean
    at previously observed values (to handle noisy observations).

    The probability of improvement can be easily computed using the
    cumulative distribution function of the standard normal distribution $`\Phi`$:
    $$`\text{PI}(x) = \Phi\left(\frac{f(x_{\text{best}}) - \mu}{\sigma}\right)`$$
    where $`\mu`$ and $`\sigma`$ are the mean and standard deviation of the
    predictive distribution of the objective function at $`x`$.

    References
    ----------
    [1] Kushner, H. J. (1964).
    A new method of locating the maximum point of an arbitrary multipeak curve in the presence of noise.
    Journal of Basic Engineering, 86(1), 97-106.

    [2] Shahriari, B., Swersky, K., Wang, Z., Adams, R. P., & de Freitas, N. (2016).
    Taking the human out of the loop: A review of Bayesian optimization.
    Proceedings of the IEEE, 104(1), 148-175. doi: 10.1109/JPROC.2015.2494218
    """

    def build_utility_function(
        self,
        posteriors: Mapping[str, ConjugatePosterior],
        datasets: Mapping[str, Dataset],
        key: KeyArray,
    ) -> SinglePointUtilityFunction:
        """
        Constructs the probability of improvement utility function
        using the predictive posterior of the objective function.

        Args:
            posteriors (Mapping[str, AbstractPosterior]): Dictionary of posteriors to be
            used to form the utility function. One of the posteriors must correspond
            to the `OBJECTIVE` key, as we sample from the objective posterior to form
            the utility function.
            datasets (Mapping[str, Dataset]): Dictionary of datasets which may be used
            to form the utility function. Keys in `datasets` should correspond to
            keys in `posteriors`. One of the datasets must correspond
            to the `OBJECTIVE` key.
            key (KeyArray): JAX PRNG key used for random number generation. Since
            the probability of improvement is computed deterministically
            from the predictive posterior, the key is not used.

        Returns:
            SinglePointUtilityFunction: the probability of improvement utility function.
        """
        self.check_objective_present(posteriors, datasets)

        objective_posterior = posteriors[OBJECTIVE]
        if not isinstance(objective_posterior, ConjugatePosterior):
            raise ValueError(
                "Objective posterior must be a ConjugatePosterior to compute the Probability of Improvement using a Gaussian CDF."
            )

        objective_dataset = datasets[OBJECTIVE]
        if (
            objective_dataset.X is None
            or objective_dataset.n == 0
            or objective_dataset.y is None
        ):
            raise ValueError(
                "Objective dataset must be non-empty to compute the "
                "Probability of Improvement (since we need a "
                "`best_y` value)."
            )

        def probability_of_improvement(x_test: Num[Array, "N D"]):
            best_y = get_best_latent_observation_val(
                objective_posterior, objective_dataset
            )
            predictive_dist = objective_posterior.predict(x_test, objective_dataset)

            normal_dist = tfp.distributions.Normal(
                loc=predictive_dist.mean(),
                scale=predictive_dist.stddev(),
            )

            return normal_dist.cdf(best_y).reshape(-1, 1)

        return probability_of_improvement



from flax import nnx
import jax.numpy as jnp
import pytest

from gpjax.parameters import (
    DEFAULT_BIJECTION,
    LowerTriangular,
    Parameter,
    PositiveReal,
    Real,
    SigmoidBounded,
    Static,
    transform,
)


@pytest.mark.parametrize(
    "param, value",
    [
        (PositiveReal, 1.0),
        (Real, 2.0),
        (SigmoidBounded, 0.5),
    ],
)
def test_transform(param, value):
    # Create mock parameters and bijectors
    params = nnx.State(
        {
            "param1": param(value),
            "param2": Parameter(2.0, tag="real"),
        }
    )

    # Test forward transformation
    t_params = transform(params, DEFAULT_BIJECTION)
    t_param1_expected = DEFAULT_BIJECTION[params["param1"]._tag].forward(value)
    assert jnp.allclose(t_params["param1"].value, t_param1_expected)
    assert jnp.allclose(t_params["param2"].value, 2.0)

    # Test inverse transformation
    it_params = transform(t_params, DEFAULT_BIJECTION, inverse=True)
    assert it_params == params


@pytest.mark.parametrize(
    "param, tag",
    [
        (PositiveReal(1.0), "positive"),
        (Real(2.0), "real"),
        (SigmoidBounded(0.5), "sigmoid"),
        (Static(2.0), "static"),
        (LowerTriangular(jnp.eye(2)), "lower_triangular"),
    ],
)
def test_default_tags(param, tag):
    assert param._tag == tag



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

import jax.numpy as jnp
import jax.random as jr
import pytest
import tensorflow_probability.substrates.jax as tfp

from gpjax.decision_making.test_functions import (
    AbstractContinuousTestFunction,
    Forrester,
    LogarithmicGoldsteinPrice,
    Quadratic,
)
from gpjax.typing import KeyArray


def test_abstract_continuous_test_function():
    with pytest.raises(TypeError):
        AbstractContinuousTestFunction()


@pytest.mark.parametrize(
    "test_function", [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()]
)
def test_minimizer_evaluates_to_minimum(test_function: AbstractContinuousTestFunction):
    evaluated_minimizer = test_function.evaluate(test_function.minimizer)
    assert evaluated_minimizer.dtype == jnp.float64
    assert jnp.allclose(
        evaluated_minimizer,
        test_function.minimum,
        atol=1e-5,
    ).all()


@pytest.mark.parametrize(
    "test_function", [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()]
)
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
def test_correct_minimum(test_function: AbstractContinuousTestFunction, key: KeyArray):
    test_x = test_function.generate_test_points(100000, key)
    function_vals = test_function.evaluate(test_x)
    vals_minus_minimum = function_vals - test_function.minimum
    assert function_vals.dtype == jnp.float64
    assert jnp.all(vals_minus_minimum >= -1e-5)


@pytest.mark.parametrize(
    "test_function", [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()]
)
def test_correct_dtypes(test_function: AbstractContinuousTestFunction):
    minimizer = test_function.minimizer
    minimum = test_function.minimum
    dataset = test_function.generate_dataset(10, jr.key(42))
    test_x = test_function.generate_test_points(10, jr.key(42))
    assert minimizer.dtype == jnp.float64
    assert minimum.dtype == jnp.float64
    assert dataset.X.dtype == jnp.float64
    assert dataset.y.dtype == jnp.float64
    assert test_x.dtype == jnp.float64


@pytest.mark.parametrize(
    "test_function, dimensionality",
    [(Quadratic(), 1), (Forrester(), 1), (LogarithmicGoldsteinPrice(), 2)],
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
def test_test_points_shape(
    test_function: AbstractContinuousTestFunction, dimensionality: int, num_samples: int
):
    test_X = test_function.generate_test_points(num_samples, jr.key(42))
    assert test_X.shape == (num_samples, dimensionality)


@pytest.mark.parametrize(
    "test_function, dimensionality",
    [(Quadratic(), 1), (Forrester(), 1), (LogarithmicGoldsteinPrice(), 2)],
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
def test_dataset_shapes(
    test_function: AbstractContinuousTestFunction, dimensionality: int, num_samples: int
):
    dataset = test_function.generate_dataset(num_samples, jr.key(42))
    assert dataset.X.shape == (num_samples, dimensionality)
    assert dataset.y.shape == (num_samples, 1)


@pytest.mark.parametrize(
    "test_function",
    [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()],
)
def test_noiseless_dataset(test_function: AbstractContinuousTestFunction):
    dataset = test_function.generate_dataset(100, jr.key(42))
    expected_y = test_function(dataset.X)
    assert jnp.all(dataset.y == expected_y)


@pytest.mark.parametrize(
    "test_function",
    [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()],
)
@pytest.mark.parametrize("obs_stddev", [1.0, 2.3])
def test_noisy_dataset(
    test_function: AbstractContinuousTestFunction, obs_stddev: float
):
    num_points = 100
    dataset = test_function.generate_dataset(num_points, jr.key(42), obs_stddev)
    noise = tfp.distributions.Normal(
        jnp.zeros(num_points), obs_stddev * jnp.ones(num_points)
    )
    expected_y = test_function(dataset.X) + jnp.transpose(
        noise.sample(sample_shape=[1], seed=jr.key(42))
    )
    assert jnp.all(dataset.y == expected_y)


@pytest.mark.parametrize(
    "test_function", [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()]
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
def test_same_key_same_dataset(
    test_function: AbstractContinuousTestFunction, num_samples: int, key: KeyArray
):
    dataset_one = test_function.generate_dataset(num_samples, key)
    dataset_two = test_function.generate_dataset(num_samples, key)
    assert jnp.equal(dataset_one.X, dataset_two.X).all()
    assert jnp.equal(dataset_one.y, dataset_two.y).all()


@pytest.mark.parametrize(
    "test_function", [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()]
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
def test_different_key_different_dataset(
    test_function: AbstractContinuousTestFunction, num_samples: int, key: KeyArray
):
    dataset_one = test_function.generate_dataset(num_samples, key)
    key, _ = jr.split(key)
    dataset_two = test_function.generate_dataset(num_samples, key)
    assert not jnp.equal(dataset_one.X, dataset_two.X).all()
    assert not jnp.equal(dataset_one.y, dataset_two.y).all()


@pytest.mark.parametrize(
    "test_function", [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()]
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
def test_same_key_same_test_points(
    test_function: AbstractContinuousTestFunction, num_samples: int, key: KeyArray
):
    test_points_one = test_function.generate_test_points(num_samples, key)
    test_points_two = test_function.generate_test_points(num_samples, key)
    assert jnp.equal(test_points_one, test_points_two).all()


@pytest.mark.parametrize(
    "test_function", [Quadratic(), Forrester(), LogarithmicGoldsteinPrice()]
)
@pytest.mark.parametrize("num_samples", [1, 10, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
def test_different_key_different_test_points(
    test_function: AbstractContinuousTestFunction, num_samples: int, key: KeyArray
):
    test_points_one = test_function.generate_test_points(num_samples, key)
    key, _ = jr.split(key)
    test_points_two = test_function.generate_test_points(num_samples, key)
    assert not jnp.equal(test_points_one, test_points_two).all()



# Enable Float64 for more stable matrix inversions.
from jax import config

config.update("jax_enable_x64", True)


import jax.numpy as jnp
from jaxtyping import (
    Array,
    Float,
)
import pytest

from gpjax.mean_functions import (
    AbstractMeanFunction,
    Constant,
    Zero,
)


def test_abstract() -> None:
    # Check abstract mean function cannot be instantiated, as the `__call__` method is not defined.
    with pytest.raises(TypeError):
        AbstractMeanFunction()

    # Check a "dummy" mean function with defined abstract method, `__call__`, can be instantiated.
    class DummyMeanFunction(AbstractMeanFunction):
        def __call__(self, x: Float[Array, " D"]) -> Float[Array, "1"]:
            return jnp.array([1.0])

    mf = DummyMeanFunction()
    assert isinstance(mf, AbstractMeanFunction)
    assert (mf(jnp.array([1.0])) == jnp.array([1.0])).all()
    assert (mf(jnp.array([2.0, 3.0])) == jnp.array([1.0])).all()


@pytest.mark.parametrize(
    "constant", [jnp.array([0.0]), jnp.array([1.0]), jnp.array([3.0])]
)
def test_constant(constant: Float[Array, " Q"]) -> None:
    mf = Constant(constant=constant)

    assert isinstance(mf, AbstractMeanFunction)
    assert (mf(jnp.array([[1.0]])) == jnp.array([constant])).all()
    assert (mf(jnp.array([[2.0, 3.0]])) == jnp.array([constant])).all()
    assert (mf(jnp.array([[1.0], [2.0]])) == jnp.array([constant, constant])).all()
    assert (
        mf(jnp.array([[1.0, 2.0], [3.0, 4.0]])) == jnp.array([constant, constant])
    ).all()


# TODO: rewrite this test after work on fit
# def test_zero_mean_remains_zero() -> None:
#     key = jr.PRNGKey(123)

#     x = jr.uniform(key=key, minval=0, maxval=1, shape=(20, 1))
#     y = jnp.full((20, 1), 50, dtype=jnp.float64)  # Dataset with non-zero mean
#     D = gpx.Dataset(X=x, y=y)

#     kernel = gpx.kernels.Constant(constant=jnp.array(0.0))
#     kernel = kernel.replace_trainable(
#         constant=False
#     )  # Prevent kernel from modelling non-zero mean
#     meanf = Zero()
#     prior = gpx.gps.Prior(mean_function=meanf, kernel=kernel)
#     likelihood = gpx.likelihoods.Gaussian(
#         num_datapoints=D.n, obs_stddev=jnp.array(1e-3)
#     )
#     likelihood = likelihood.replace_trainable(obs_stddev=False)
#     posterior = prior * likelihood

#     negative_mll = gpx.objectives.ConjugateMLL(negative=True)
#     opt_posterior, _ = gpx.fit(
#         model=posterior,
#         objective=negative_mll,
#         train_data=D,
#         optim=ox.adam(learning_rate=0.5),
#         num_iters=1000,
#         safe=True,
#         key=key,
#     )

#     assert opt_posterior.prior.mean_function.constant == 0.0


def test_initialising_zero_mean_with_constant_raises_error():
    with pytest.raises(TypeError):
        Zero(constant=jnp.array([1.0]))



import typing as tp

from flax import nnx
import jax.numpy as jnp
import jax.tree_util as jtu
from jax.typing import ArrayLike
import tensorflow_probability.substrates.jax.bijectors as tfb

T = tp.TypeVar("T", bound=tp.Union[ArrayLike, list[float]])
ParameterTag = str


def transform(
    params: nnx.State,
    params_bijection: tp.Dict[str, tfb.Bijector],
    inverse: bool = False,
) -> nnx.State:
    r"""Transforms parameters using a bijector.

    Example:
    ```pycon
        >>> from gpjax.parameters import PositiveReal, transform
        >>> import jax.numpy as jnp
        >>> import tensorflow_probability.substrates.jax.bijectors as tfb
        >>> from flax import nnx
        >>> params = nnx.State(
        >>>     {
        >>>         "a": PositiveReal(jnp.array([1.0])),
        >>>         "b": PositiveReal(jnp.array([2.0])),
        >>>     }
        >>> )
        >>> params_bijection = {'positive': tfb.Softplus()}
        >>> transformed_params = transform(params, params_bijection)
        >>> print(transformed_params["a"].value)
         [1.3132617]
    ```


    Args:
        params: A nnx.State object containing parameters to be transformed.
        params_bijection: A dictionary mapping parameter types to bijectors.
        inverse: Whether to apply the inverse transformation.

    Returns:
        State: A new nnx.State object containing the transformed parameters.
    """

    def _inner(param):
        bijector = params_bijection.get(param._tag, tfb.Identity())
        if inverse:
            transformed_value = bijector.inverse(param.value)
        else:
            transformed_value = bijector.forward(param.value)

        param = param.replace(transformed_value)
        return param

    gp_params, *other_params = params.split(Parameter, ...)

    transformed_gp_params: nnx.State = jtu.tree_map(
        lambda x: _inner(x),
        gp_params,
        is_leaf=lambda x: isinstance(x, nnx.VariableState),
    )
    return nnx.State.merge(transformed_gp_params, *other_params)


class Parameter(nnx.Variable[T]):
    """Parameter base class.

    All trainable parameters in GPJax should inherit from this class.

    """

    def __init__(self, value: T, tag: ParameterTag, **kwargs):
        _check_is_arraylike(value)

        super().__init__(value=jnp.asarray(value), **kwargs)
        self._tag = tag


class PositiveReal(Parameter[T]):
    """Parameter that is strictly positive."""

    def __init__(self, value: T, tag: ParameterTag = "positive", **kwargs):
        super().__init__(value=value, tag=tag, **kwargs)

        _check_is_positive(self.value)


class Real(Parameter[T]):
    """Parameter that can take any real value."""

    def __init__(self, value: T, tag: ParameterTag = "real", **kwargs):
        super().__init__(value, tag, **kwargs)


class SigmoidBounded(Parameter[T]):
    """Parameter that is bounded between 0 and 1."""

    def __init__(self, value: T, tag: ParameterTag = "sigmoid", **kwargs):
        super().__init__(value=value, tag=tag, **kwargs)

        _check_in_bounds(self.value, 0.0, 1.0)


class Static(nnx.Variable[T]):
    """Static parameter that is not trainable."""

    def __init__(self, value: T, tag: ParameterTag = "static", **kwargs):
        _check_is_arraylike(value)

        super().__init__(value=jnp.asarray(value), tag=tag, **kwargs)
        self._tag = tag


class LowerTriangular(Parameter[T]):
    """Parameter that is a lower triangular matrix."""

    def __init__(self, value: T, tag: ParameterTag = "lower_triangular", **kwargs):
        super().__init__(value=value, tag=tag, **kwargs)

        _check_is_square(self.value)
        _check_is_lower_triangular(self.value)


DEFAULT_BIJECTION = {
    "positive": tfb.Softplus(),
    "real": tfb.Identity(),
    "sigmoid": tfb.Sigmoid(low=0.0, high=1.0),
    "lower_triangular": tfb.FillTriangular(),
}


def _check_is_arraylike(value: T):
    if not isinstance(value, (ArrayLike, list)):
        raise TypeError(
            f"Expected parameter value to be an array-like type. Got {value}."
        )


def _check_is_positive(value: T):
    if jnp.any(value < 0):
        raise ValueError(
            f"Expected parameter value to be strictly positive. Got {value}."
        )


def _check_is_square(value: T):
    if value.shape[0] != value.shape[1]:
        raise ValueError(
            f"Expected parameter value to be a square matrix. Got {value}."
        )


def _check_is_lower_triangular(value: T):
    if not jnp.all(jnp.tril(value) == value):
        raise ValueError(
            f"Expected parameter value to be a lower triangular matrix. Got {value}."
        )


def _check_in_bounds(value: T, low: float, high: float):
    if jnp.any((value < low) | (value > high)):
        raise ValueError(
            f"Expected parameter value to be bounded between {low} and {high}. Got {value}."
        )



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

from beartype.typing import (
    Callable,
    Union,
)
import jax.numpy as jnp
import jax.random as jr
import optax as ox
import pytest

from gpjax.decision_making.posterior_handler import PosteriorHandler
from gpjax.decision_making.test_functions import (
    Forrester,
    PoissonTestFunction,
)
from gpjax.gps import Prior
from gpjax.kernels import Matern52
from gpjax.likelihoods import (
    AbstractLikelihood,
    Gaussian,
    Poisson,
)
from gpjax.mean_functions import Constant
from gpjax.objectives import (
    Objective,
    conjugate_mll,
    non_conjugate_mll,
)


def gaussian_likelihood_builder(num_datapoints: int) -> Gaussian:
    return Gaussian(num_datapoints=num_datapoints)


def poisson_likelihood_builder(num_datapoints: int) -> Poisson:
    return Poisson(num_datapoints=num_datapoints)


@pytest.mark.parametrize("num_optimization_iters", [0, -1, -10])
def test_posterior_handler_erroneous_num_optimization_iterations_raises_error(
    num_optimization_iters: int,
):
    mean_function = Constant()
    kernel = Matern52()
    prior = Prior(mean_function=mean_function, kernel=kernel)
    likelihood_builder = gaussian_likelihood_builder
    with pytest.raises(ValueError):
        PosteriorHandler(
            prior=prior,
            likelihood_builder=likelihood_builder,
            optimization_objective=conjugate_mll,
            optimizer=ox.adam(learning_rate=0.01),
            num_optimization_iters=num_optimization_iters,
        )


def test_get_optimized_posterior_with_no_key_raises_error():
    mean_function = Constant()
    kernel = Matern52()
    prior = Prior(mean_function=mean_function, kernel=kernel)
    likelihood_builder = gaussian_likelihood_builder
    posterior_handler = PosteriorHandler(
        prior=prior,
        likelihood_builder=likelihood_builder,
        optimization_objective=conjugate_mll,
        optimizer=ox.adam(learning_rate=0.01),
        num_optimization_iters=10,
    )
    toy_function = Forrester()
    dataset = toy_function.generate_dataset(num_points=5, key=jr.key(42))
    with pytest.raises(ValueError):
        posterior_handler.get_posterior(dataset=dataset, optimize=True)


def test_update_and_optimize_posterior_with_no_key_raises_error():
    mean_function = Constant()
    kernel = Matern52()
    prior = Prior(mean_function=mean_function, kernel=kernel)
    likelihood_builder = gaussian_likelihood_builder
    posterior_handler = PosteriorHandler(
        prior=prior,
        likelihood_builder=likelihood_builder,
        optimization_objective=conjugate_mll,
        optimizer=ox.adam(learning_rate=0.01),
        num_optimization_iters=10,
    )
    toy_function = Forrester()
    dataset = toy_function.generate_dataset(num_points=5, key=jr.key(42))
    initial_posterior = posterior_handler.get_posterior(dataset=dataset, optimize=False)
    with pytest.raises(ValueError):
        posterior_handler.update_posterior(
            dataset=dataset, previous_posterior=initial_posterior, optimize=True
        )


@pytest.mark.parametrize("num_datapoints", [1, 50])
@pytest.mark.parametrize(
    "likelihood_builder, training_objective, test_function",
    [
        (gaussian_likelihood_builder, conjugate_mll, Forrester()),
        (
            poisson_likelihood_builder,
            non_conjugate_mll,
            PoissonTestFunction(),
        ),
    ],
)
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_get_posterior_no_optimization_correct_num_datapoints_and_not_optimized(
    num_datapoints: int,
    likelihood_builder: Callable[[int], AbstractLikelihood],
    training_objective: Objective,
    test_function: Union[Forrester, PoissonTestFunction],
):
    mean_function = Constant(constant=jnp.array([1.0]))
    kernel = Matern52(lengthscale=jnp.array([0.5]), variance=jnp.array(1.0))
    prior = Prior(mean_function=mean_function, kernel=kernel)
    posterior_handler = PosteriorHandler(
        prior=prior,
        likelihood_builder=likelihood_builder,
        optimization_objective=training_objective,
        optimizer=ox.adam(learning_rate=0.01),
        num_optimization_iters=10,
    )
    dataset = test_function.generate_dataset(num_points=num_datapoints, key=jr.key(42))
    posterior = posterior_handler.get_posterior(dataset=dataset, optimize=False)
    assert posterior.likelihood.num_datapoints == num_datapoints
    assert posterior.prior.mean_function.constant.value == jnp.array([1.0])
    assert posterior.prior.kernel.lengthscale.value == jnp.array([0.5])
    assert posterior.prior.kernel.variance.value == jnp.array(1.0)


@pytest.mark.parametrize("num_datapoints", [5, 50])
@pytest.mark.parametrize(
    "likelihood_builder, training_objective, test_function",
    [
        (gaussian_likelihood_builder, conjugate_mll, Forrester()),
        (
            poisson_likelihood_builder,
            non_conjugate_mll,
            PoissonTestFunction(),
        ),
    ],
)
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_get_posterior_with_optimization_correct_num_datapoints_and_optimized(
    num_datapoints: int,
    likelihood_builder: Callable[[int], AbstractLikelihood],
    training_objective: Objective,
    test_function: Union[Forrester, PoissonTestFunction],
):
    mean_function = Constant(constant=jnp.array([1.0]))
    kernel = Matern52(lengthscale=jnp.array([0.5]), variance=jnp.array(1.0))
    prior = Prior(mean_function=mean_function, kernel=kernel)
    non_optimized_posterior = prior * likelihood_builder(num_datapoints)
    posterior_handler = PosteriorHandler(
        prior=prior,
        likelihood_builder=likelihood_builder,
        optimization_objective=training_objective,
        optimizer=ox.adam(learning_rate=0.01),
        num_optimization_iters=10,
    )
    dataset = test_function.generate_dataset(num_points=num_datapoints, key=jr.key(42))
    optimized_posterior = posterior_handler.get_posterior(
        dataset=dataset, optimize=True, key=jr.key(42)
    )
    assert optimized_posterior.likelihood.num_datapoints == num_datapoints
    assert optimized_posterior.prior.mean_function.constant != jnp.array([1.0])
    assert optimized_posterior.prior.kernel.lengthscale != jnp.array([0.5])
    assert optimized_posterior.prior.kernel.variance != jnp.array(1.0)
    assert training_objective(optimized_posterior, dataset) < training_objective(
        non_optimized_posterior, dataset
    )  # Ensure optimization reduces training objective


@pytest.mark.parametrize("initial_num_datapoints", [5, 50])
@pytest.mark.parametrize(
    "likelihood_builder, training_objective, test_function",
    [
        (gaussian_likelihood_builder, conjugate_mll, Forrester()),
        (
            poisson_likelihood_builder,
            non_conjugate_mll,
            PoissonTestFunction(),
        ),
    ],
)
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_update_posterior_no_optimize_same_prior_parameters_and_different_num_datapoints(
    initial_num_datapoints: int,
    likelihood_builder: Callable[[int], AbstractLikelihood],
    training_objective: Objective,
    test_function: Union[Forrester, PoissonTestFunction],
):
    mean_function = Constant(constant=jnp.array([1.0]))
    kernel = Matern52(lengthscale=jnp.array([0.5]), variance=jnp.array(1.0))
    prior = Prior(mean_function=mean_function, kernel=kernel)
    posterior_handler = PosteriorHandler(
        prior=prior,
        likelihood_builder=likelihood_builder,
        optimization_objective=training_objective,
        optimizer=ox.adam(learning_rate=0.01),
        num_optimization_iters=10,
    )
    initial_dataset = test_function.generate_dataset(
        num_points=initial_num_datapoints, key=jr.key(42)
    )
    initial_posterior = posterior_handler.get_posterior(
        dataset=initial_dataset, optimize=False
    )
    updated_dataset = initial_dataset + test_function.generate_dataset(
        num_points=1, key=jr.key(42)
    )
    assert updated_dataset.n == initial_dataset.n + 1
    updated_posterior = posterior_handler.update_posterior(
        dataset=updated_dataset, previous_posterior=initial_posterior, optimize=False
    )
    assert (
        updated_posterior.prior.kernel.lengthscale
        == initial_posterior.prior.kernel.lengthscale
    )
    assert (
        updated_posterior.prior.kernel.variance
        == initial_posterior.prior.kernel.variance
    )
    assert (
        updated_posterior.prior.mean_function.constant
        == initial_posterior.prior.mean_function.constant
    )
    assert updated_posterior.likelihood.num_datapoints == updated_dataset.n


@pytest.mark.parametrize("initial_num_datapoints", [5, 50])
@pytest.mark.parametrize(
    "likelihood_builder, training_objective, test_function",
    [
        (gaussian_likelihood_builder, conjugate_mll, Forrester()),
        (
            poisson_likelihood_builder,
            non_conjugate_mll,
            PoissonTestFunction(),
        ),
    ],
)
@pytest.mark.filterwarnings("ignore:y is not of type float64")
def test_update_posterior_with_optimization_updated_prior_parameters_and_different_num_datapoints(
    initial_num_datapoints: int,
    likelihood_builder: Callable[[int], AbstractLikelihood],
    training_objective: Objective,
    test_function: Union[Forrester, PoissonTestFunction],
):
    mean_function = Constant(constant=jnp.array([1.0]))
    kernel = Matern52(lengthscale=jnp.array([0.5]), variance=jnp.array(1.0))
    prior = Prior(mean_function=mean_function, kernel=kernel)
    posterior_handler = PosteriorHandler(
        prior=prior,
        likelihood_builder=likelihood_builder,
        optimization_objective=training_objective,
        optimizer=ox.adam(learning_rate=0.01),
        num_optimization_iters=10,
    )
    initial_dataset = test_function.generate_dataset(
        num_points=initial_num_datapoints, key=jr.key(42)
    )
    initial_posterior = posterior_handler.get_posterior(
        dataset=initial_dataset, optimize=False
    )
    updated_dataset = initial_dataset + test_function.generate_dataset(
        num_points=1, key=jr.key(42)
    )
    assert updated_dataset.n == initial_dataset.n + 1
    non_optimized_updated_posterior = posterior_handler.update_posterior(
        dataset=updated_dataset, previous_posterior=initial_posterior, optimize=False
    )
    optimized_updated_posterior = posterior_handler.update_posterior(
        dataset=updated_dataset,
        previous_posterior=initial_posterior,
        optimize=True,
        key=jr.key(42),
    )
    assert (
        optimized_updated_posterior.prior.kernel.lengthscale
        != initial_posterior.prior.kernel.lengthscale
    )
    assert (
        optimized_updated_posterior.prior.kernel.variance
        != initial_posterior.prior.kernel.variance
    )
    assert (
        optimized_updated_posterior.prior.mean_function.constant
        != initial_posterior.prior.mean_function.constant
    )
    assert optimized_updated_posterior.likelihood.num_datapoints == updated_dataset.n
    assert training_objective(
        optimized_updated_posterior, updated_dataset
    ) < training_objective(
        non_optimized_updated_posterior, updated_dataset
    )  # Ensure optimization reduces training objective



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

from gpjax.decision_making.utility_functions.expected_improvement import (
    ExpectedImprovement,
)

config.update("jax_enable_x64", True)

from beartype.typing import Type
import jax.random as jr
from jaxtyping import TypeCheckError
import pytest

from gpjax.decision_making.test_functions.continuous_functions import (
    AbstractContinuousTestFunction,
    Forrester,
    LogarithmicGoldsteinPrice,
)
from gpjax.decision_making.utility_functions.base import (
    AbstractSinglePointUtilityFunctionBuilder,
)
from gpjax.decision_making.utility_functions.probability_of_improvement import (
    ProbabilityOfImprovement,
)
from gpjax.decision_making.utility_functions.thompson_sampling import ThompsonSampling
from gpjax.decision_making.utils import OBJECTIVE
from gpjax.typing import KeyArray
from tests.test_decision_making.utils import (
    generate_dummy_conjugate_posterior,
    generate_dummy_non_conjugate_posterior,
)


@pytest.mark.filterwarnings(
    "ignore::UserWarning"
)  # Sampling with tfp causes JAX to raise a UserWarning due to some internal logic around jnp.argsort
@pytest.mark.parametrize(
    "utility_function_builder, utility_function_kwargs",
    [
        (ExpectedImprovement, {}),
        (ProbabilityOfImprovement, {}),
        (ThompsonSampling, {"num_features": 100}),
    ],
)
def test_utility_function_no_objective_posterior_raises_error(
    utility_function_builder: Type[AbstractSinglePointUtilityFunctionBuilder],
    utility_function_kwargs: dict,
):
    key = jr.key(42)
    forrester = Forrester()
    dataset = forrester.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset)
    posteriors = {"CONSTRAINT": posterior}
    datasets = {OBJECTIVE: dataset}
    with pytest.raises(ValueError):
        utility_function = utility_function_builder(**utility_function_kwargs)
        utility_function.build_utility_function(
            posteriors=posteriors, datasets=datasets, key=key
        )


@pytest.mark.filterwarnings(
    "ignore::UserWarning"
)  # Sampling with tfp causes JAX to raise a UserWarning due to some internal logic around jnp.argsort
@pytest.mark.parametrize(
    "utility_function_builder, utility_function_kwargs",
    [
        (ExpectedImprovement, {}),
        (ProbabilityOfImprovement, {}),
        (ThompsonSampling, {"num_features": 100}),
    ],
)
def test_utility_function_no_objective_dataset_raises_error(
    utility_function_builder: Type[AbstractSinglePointUtilityFunctionBuilder],
    utility_function_kwargs: dict,
):
    key = jr.key(42)
    forrester = Forrester()
    dataset = forrester.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset)
    posteriors = {OBJECTIVE: posterior}
    datasets = {"CONSTRAINT": dataset}
    with pytest.raises(ValueError):
        utility_function = utility_function_builder(**utility_function_kwargs)
        utility_function.build_utility_function(
            posteriors=posteriors, datasets=datasets, key=key
        )


@pytest.mark.filterwarnings(
    "ignore::UserWarning"
)  # Sampling with tfp causes JAX to raise a UserWarning due to some internal logic around jnp.argsort
@pytest.mark.parametrize(
    "utility_function_builder, utility_function_kwargs",
    [
        (ExpectedImprovement, {}),
        (ProbabilityOfImprovement, {}),
        (ThompsonSampling, {"num_features": 100}),
    ],
)
def test_non_conjugate_posterior_raises_error(
    utility_function_builder: Type[AbstractSinglePointUtilityFunctionBuilder],
    utility_function_kwargs: dict,
):
    key = jr.key(42)
    forrester = Forrester()
    dataset = forrester.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_non_conjugate_posterior(dataset)
    posteriors = {OBJECTIVE: posterior}
    datasets = {OBJECTIVE: dataset}
    with pytest.raises(TypeCheckError):
        utility_function = utility_function_builder(**utility_function_kwargs)
        utility_function.build_utility_function(
            posteriors=posteriors, datasets=datasets, key=key
        )


@pytest.mark.parametrize(
    "utility_function_builder, utility_function_kwargs",
    [
        (ExpectedImprovement, {}),
        (ProbabilityOfImprovement, {}),
        (ThompsonSampling, {"num_features": 100}),
    ],
)
@pytest.mark.parametrize(
    "test_target_function",
    [(Forrester()), (LogarithmicGoldsteinPrice())],
)
@pytest.mark.parametrize("num_test_points", [50, 100])
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.filterwarnings(
    "ignore::UserWarning"
)  # Sampling with tfp causes JAX to raise a UserWarning due to some internal logic around jnp.argsort
def test_utility_functions_have_correct_shapes(
    utility_function_builder: Type[AbstractSinglePointUtilityFunctionBuilder],
    utility_function_kwargs: dict,
    test_target_function: AbstractContinuousTestFunction,
    num_test_points: int,
    key: KeyArray,
):
    dataset = test_target_function.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset)
    posteriors = {OBJECTIVE: posterior}
    datasets = {OBJECTIVE: dataset}
    utility_builder = utility_function_builder(**utility_function_kwargs)
    utility_function = utility_builder.build_utility_function(
        posteriors=posteriors, datasets=datasets, key=key
    )
    test_key, _ = jr.split(key)
    test_X = test_target_function.generate_test_points(num_test_points, test_key)
    utility_function_values = utility_function(test_X)
    assert utility_function_values.shape == (num_test_points, 1)



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from beartype.roar import BeartypeCallHintParamViolation
from jax import config
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Array,
    Float,
)
import pytest

from gpjax.decision_making.search_space import (
    AbstractSearchSpace,
    ContinuousSearchSpace,
)

config.update("jax_enable_x64", True)


def test_abstract_search_space():
    with pytest.raises(TypeError):
        AbstractSearchSpace()


def test_continuous_search_space_empty_bounds():
    with pytest.raises(ValueError):
        ContinuousSearchSpace(lower_bounds=jnp.array([]), upper_bounds=jnp.array([]))


@pytest.mark.parametrize(
    "lower_bounds, upper_bounds",
    [
        (jnp.array([0.0], dtype=jnp.float64), jnp.array([1.0], jnp.float32)),
        (jnp.array([0.0], dtype=jnp.float32), jnp.array([1.0], jnp.float64)),
    ],
)
def test_continuous_search_space_dtype_consistency(
    lower_bounds: Float[Array, " D"], upper_bounds: Float[Array, " D"]
):
    with pytest.raises(ValueError):
        ContinuousSearchSpace(lower_bounds=lower_bounds, upper_bounds=upper_bounds)


@pytest.mark.parametrize(
    "lower_bounds, upper_bounds",
    [
        (jnp.array([0.0]), jnp.array([1.0, 1.0])),
        (jnp.array([0.0, 0.0]), jnp.array([1.0])),
    ],
)
def test_continous_search_space_bounds_shape_consistency(
    lower_bounds: Float[Array, " D1"], upper_bounds: Float[Array, " D2"]
):
    with pytest.raises((BeartypeCallHintParamViolation, ValueError)):
        ContinuousSearchSpace(lower_bounds=lower_bounds, upper_bounds=upper_bounds)


@pytest.mark.parametrize(
    "lower_bounds, upper_bounds",
    [
        (jnp.array([1.0]), jnp.array([0.0])),
        (jnp.array([1.0, 1.0]), jnp.array([0.0, 2.0])),
        (jnp.array([1.0, 1.0]), jnp.array([2.0, 0.0])),
    ],
)
def test_continuous_search_space_bounds_values_consistency(
    lower_bounds: Float[Array, " D"], upper_bounds: Float[Array, " D"]
):
    with pytest.raises(ValueError):
        ContinuousSearchSpace(lower_bounds=lower_bounds, upper_bounds=upper_bounds)


@pytest.mark.parametrize(
    "continuous_search_space, dimensionality",
    [
        (ContinuousSearchSpace(jnp.array([0.0]), jnp.array([1.0])), 1),
        (ContinuousSearchSpace(jnp.array([0.0, 0.0]), jnp.array([1.0, 1.0])), 2),
        (
            ContinuousSearchSpace(
                jnp.array([0.0, 0.0, 0.0]), jnp.array([1.0, 1.0, 1.0])
            ),
            3,
        ),
    ],
)
def test_continuous_search_space_dimensionality(
    continuous_search_space: ContinuousSearchSpace, dimensionality: int
):
    assert continuous_search_space.dimensionality == dimensionality


@pytest.mark.parametrize(
    "continuous_search_space",
    [
        ContinuousSearchSpace(jnp.array([0.0]), jnp.array([1.0])),
        ContinuousSearchSpace(jnp.array([0.0, 0.0]), jnp.array([1.0, 1.0])),
        ContinuousSearchSpace(jnp.array([0.0, 0.0, 0.0]), jnp.array([1.0, 1.0, 1.0])),
    ],
)
@pytest.mark.parametrize("num_points", [0, -1])
def test_continous_search_space_invalid_sample_num_points(
    continuous_search_space: ContinuousSearchSpace, num_points: int
):
    with pytest.raises(ValueError):
        continuous_search_space.sample(num_points=num_points, key=jr.key(42))


@pytest.mark.parametrize(
    "continuous_search_space, dimensionality",
    [
        (ContinuousSearchSpace(jnp.array([0.0]), jnp.array([1.0])), 1),
        (ContinuousSearchSpace(jnp.array([0.0, 0.0]), jnp.array([1.0, 1.0])), 2),
        (
            ContinuousSearchSpace(
                jnp.array([0.0, 0.0, 0.0]), jnp.array([1.0, 1.0, 1.0])
            ),
            3,
        ),
    ],
)
@pytest.mark.parametrize("num_points", [1, 5, 50])
def test_continuous_search_space_sample_shape(
    continuous_search_space: ContinuousSearchSpace, dimensionality: int, num_points: int
):
    samples = continuous_search_space.sample(num_points=num_points, key=jr.key(42))
    assert samples.shape[0] == num_points
    assert samples.shape[1] == dimensionality


@pytest.mark.parametrize(
    "continuous_search_space",
    [
        ContinuousSearchSpace(jnp.array([0.0]), jnp.array([1.0])),
        ContinuousSearchSpace(jnp.array([0.0, 0.0]), jnp.array([1.0, 1.0])),
        ContinuousSearchSpace(jnp.array([0.0, 0.0, 0.0]), jnp.array([1.0, 1.0, 1.0])),
    ],
)
@pytest.mark.parametrize("key", [jr.key(42), jr.key(5)])
def test_continous_search_space_sample_same_key_same_samples(
    continuous_search_space: ContinuousSearchSpace, key: jr.key
):
    sample_one = continuous_search_space.sample(num_points=100, key=key)
    sample_two = continuous_search_space.sample(num_points=100, key=key)
    assert jnp.array_equal(sample_one, sample_two)


@pytest.mark.parametrize(
    "continuous_search_space",
    [
        ContinuousSearchSpace(jnp.array([0.0]), jnp.array([1.0])),
        ContinuousSearchSpace(jnp.array([0.0, 0.0]), jnp.array([1.0, 1.0])),
        ContinuousSearchSpace(jnp.array([0.0, 0.0, 0.0]), jnp.array([1.0, 1.0, 1.0])),
    ],
)
@pytest.mark.parametrize(
    "key_one, key_two",
    [(jr.key(42), jr.key(5)), (jr.key(1), jr.key(2))],
)
def test_continuous_search_space_different_keys_different_samples(
    continuous_search_space: ContinuousSearchSpace,
    key_one: jr.key,
    key_two: jr.key,
):
    sample_one = continuous_search_space.sample(num_points=100, key=key_one)
    sample_two = continuous_search_space.sample(num_points=100, key=key_two)
    assert not jnp.array_equal(sample_one, sample_two)


@pytest.mark.parametrize(
    "continuous_search_space",
    [
        ContinuousSearchSpace(
            lower_bounds=jnp.array([0.0]), upper_bounds=jnp.array([1.0])
        ),
        ContinuousSearchSpace(
            lower_bounds=jnp.array([0.0, 0.0]), upper_bounds=jnp.array([1.0, 2.0])
        ),
        ContinuousSearchSpace(
            lower_bounds=jnp.array([0.0, 1.0]), upper_bounds=jnp.array([2.0, 2.0])
        ),
        ContinuousSearchSpace(
            lower_bounds=jnp.array([2.4, 1.7, 4.9]),
            upper_bounds=jnp.array([5.6, 1.8, 6.0]),
        ),
    ],
)
def test_continuous_search_space_valid_sample_ranges(
    continuous_search_space: ContinuousSearchSpace,
):
    samples = continuous_search_space.sample(num_points=100, key=jr.key(42))
    for i in range(continuous_search_space.dimensionality):
        assert jnp.all(samples[:, i] >= continuous_search_space.lower_bounds[i])
        assert jnp.all(samples[:, i] <= continuous_search_space.upper_bounds[i])



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Introduction to Decision Making with GPJax
#
# In this notebook we provide an introduction to the decision making module of GPJax,
# which can be used to solve sequential decision making problems. Common examples of
# such problems include Bayesian optimisation (BO) and experimental design. For an
# in-depth introduction to Bayesian optimisation itself, be sure to checkout out our
# [Introduction to BO
# Notebook](https://docs.jaxgaussianprocesses.com/examples/bayesian_optimisation/).
#
# We'll be using BO as a case study to demonstrate how one may use the decision making
# module to solve sequential decision making problems. The goal of the decision making
# module is to provide a set of tools that can easily be used to solve a wide range of
# sequential decision making problems. The module is designed to be modular, and so it is
# easy to swap out different components of the decision making pipeline. Whilst it
# provides the functionality for quickly implementing a typical deicision making loop out
# of the box, we also hope that it will provide sufficient flexibility to allow users to
# define their own, more exotic, decision making pipelines.

# %%
# Enable Float64 for more stable matrix inversions.
from jax import config
import jax.numpy as jnp
import jax.random as jr
import matplotlib as mpl
import matplotlib.pyplot as plt
import optax as ox

from examples.utils import use_mpl_style
import gpjax as gpx
from gpjax.decision_making.decision_maker import UtilityDrivenDecisionMaker
from gpjax.decision_making.posterior_handler import PosteriorHandler
from gpjax.decision_making.search_space import ContinuousSearchSpace
from gpjax.decision_making.utility_functions import ThompsonSampling
from gpjax.decision_making.utility_maximizer import (
    ContinuousSinglePointUtilityMaximizer,
)
from gpjax.decision_making.utils import (
    OBJECTIVE,
    build_function_evaluator,
)
from gpjax.typing import (
    Array,
    Float,
)

config.update("jax_enable_x64", True)


# set the default style for plotting
use_mpl_style()

key = jr.key(42)


cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]


# %% [markdown]
# ## The Black-Box Objective Function
#
# We'll be using the same problem as in the [Introduction to BO
# Notebook](https://docs.jaxgaussianprocesses.com/examples/bayesian_optimisation/), but
# rather than focussing on the mechanics of BO we'll be looking at how one may use the
# abstractions provided by the decision making module to implement the BO loop.
#
# In BO, and sequential decision making in general, we will often have a black-box
# function of interest which we can evaluate. In this notebook we'll be using the
# Forrester function as our objective to minimise:
#
# $$f(x) = (6x - 2)^2\sin(12x-4)$$


# %%
def forrester(x: Float[Array, "N 1"]) -> Float[Array, "N 1"]:
    return (6 * x - 2) ** 2 * jnp.sin(12 * x - 4)


# %% [markdown]
# Within the decision making loop, we'll be querying the black-box objective function many
# times, and will often use the observed values to fit some probabilistic model. Thereore,
# it would be useful to have some method to which we can pass a set of points which we
# wish to query the black-box function at, and which will return a GPJax `Dataset` object
# containing the observations. We can use the `build_function_evaluator` function provided
# in `decision_making.utils` to do this. This function takes as input a dictionary of
# labelled black-box functions, and will return a function evaluator, which can be called
# with a set of points to evaluate the black-box functions at. The function evaluator will
# return a dictionary of labelled `Dataset` objects containing the observations. Note that
# in our case we only have one black-box function of interest, but in general we may have
# multiple different black-box functions, such as if we also have constraint functions.
# The use of the labels inside the dictionary returned by the function evaluator enables
# us to easily distinguish between these different observations.

# %%
function_evaluator = build_function_evaluator({OBJECTIVE: forrester})

# %% [markdown]
# ## The Search Space
#
# Having defined a method for evaluating the black-box function, we now need to define the
# search space over which we wish to optimise. In this case we'll be optimising over the
# interval $[0, 1]$. We can use the `ContinuousSearchSpace` class provided in
# `decision_making.search_space` to define this search space, as seen below:

# %%
lower_bounds = jnp.array([0.0])
upper_bounds = jnp.array([1.0])
search_space = ContinuousSearchSpace(
    lower_bounds=lower_bounds, upper_bounds=upper_bounds
)

# %% [markdown]
# The `ContinuousSearchSpace` class defines a `sample` method, which can be used to
# sample points from the search space using a space-filling design, in this case using the
# [Halton sequence](https://en.wikipedia.org/wiki/Halton_sequence). This will be useful at
# many points throughout the decision making loop, but for now let's use it to create an
# initial set of points which we can use to fit our models:

# %%
initial_x = search_space.sample(5, key)
initial_datasets = function_evaluator(initial_x)

# %% [markdown]
# ## The Surrogate Models

# %% [markdown]
# Many sequential decision making algorithms are described as being *model-based*. With
# these algorithms, we use a probabilistic model, or multiple models, to drive the
# decision making process. In ordinary BO, a probabilistic model is used to model the
# objective function, and it is updated based on observations from the black-box objective
# function. These models are often referred to as *surrogate models*, and are used to
# approximate the functions of interest. We'll be using the Gaussian process functionality
# provided by GPJax to define our surrogate models, with some wrappers provided by the
# `decision_making` module to make it easier to use these models within the decision
# making loop. We can proceed as usual when defining our priors, choosing a suitable
# mean function and kernel for the job at hand:

# %%
mean = gpx.mean_functions.Zero()
kernel = gpx.kernels.Matern52(n_dims=1)
prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)

# %% [markdown]
# One difference from GPJax is the way in which we define our likelihood. In GPJax, we
# construct our GP posteriors by defining a `likelihood` object and then multiplying it
# with our prior to get the posterior, `posterior = likelihood * prior`. However, the
# `AbstractLikelihood` objects takes `num_datapoints` as one of its arguments, and this is
# going to be changing in the case of BO, and decision making in general, as we keep
# updating our models having observed new data! In order to deal with this we'll define a
# `likelihood_builder`, which takes as an argument the number of datapoints used to
# condition our prior on, and returns a `likelihood` object. Below we use this to
# construct a `likelihood_builder` which will return a `Gaussian` likelihood, initialised
# with the correct number of datapoints:

# %%
likelihood_builder = lambda n: gpx.likelihoods.Gaussian(
    num_datapoints=n, obs_stddev=jnp.array(1e-3)
)

# %% [markdown]
# Now we have all the components required for constructing our GP posterior. Since we'll
# be updating the posterior throughout the decision making loop as we observe more data,
# it would be useful to have an object which can handle all this logic for us.
# Fortunately, the `decision_making` module provides the `PosteriorHandler` class to do
# this for us. This class takes as input a `prior` and `likeligood_builder`, which we have
# defined above. We tend to also optimise the hyperparameters of the GP prior when
# "fitting" our GP, as demonstrated in the [Regression
# notebook](https://docs.jaxgaussianprocesses.com/examples/regression/). This will be
# using the GPJax `fit` method under the hood, which requires an `optimization_objective`,
# `optimizer` and `num_optimization_iters`. Therefore, we also pass these to the
# `PosteriorHandler` as demonstrated below:

# %%
posterior_handler = PosteriorHandler(
    prior,
    likelihood_builder=likelihood_builder,
    optimization_objective=lambda p, d: -gpx.objectives.conjugate_mll(p, d),
    optimizer=ox.adam(learning_rate=0.01),
    num_optimization_iters=1000,
)
posterior_handlers = {OBJECTIVE: posterior_handler}

# %% [markdown]
# Note that we also create a labelled dictionary of `posterior_handlers`. This is a
# recurring theme with the decision making logic; we can have dictionaries containing
# datasets, posteriors and black box functions, and use labels to identify corresponding
# objects the dictionaries. For instance, here we have an "OBJECTIVE" posterior handler
# which is updated using the data in the "OBJECTIVE" dataset, which is in turn generated by the "OBJECTIVE" black-box function.
#
# Now, as the decision making loop progresses, we can use the `update_posterior` method of
# the `PosteriorHandler` to update our posterior as we observe more data. Note that we use
# the term *posterior* to refer to our GP posterior surrogate models in order to be
# consistent with the syntax used by GPJax. However, these GP posteriors are more widely
# referred to as *models* in the model-based decision making literature.

# %% [markdown]
# ## The Utility Function
#
# Now all that remains for us to define is the utiliy function, and a way of maximising
# it. Within the utility-driven decision making framework, we define a utility function,
# often using our GP surrogates, which characterises the *utility*, or *usefulness*, of
# querying the black-box function at any point within the domain of interest. We can then
# *maximise* this function to decide which point to query next. In this case we'll be
# using Thompson sampling as a utility function for determining where to query next. With
# this function we simply draw a sample from the GP posterior, and choose the minimizer
# of the sample as the point to query next. In the `decision_making` framework we create
# `UtilityFunctionBuilder` objects. Currently, we only support
# `SinglePointUtilityFunction`s, which are utility functions which characterise the
# utility of querying a single point. Thompson sampling is somewhat of a special case, as
# we can draw $B$ independent samples from the GP posterior and optimise each of these
# samples in order to obtain a *batch* of points to query next. We'll see an example of
# this later on.
#
# Within the `ThompsonSampling` utility function builder class we implement the
# `build_utility_function` method, which takes as input a dictionary containing lablled GP
# posteriors, as well as the corresponding datasets for these posteriors, and draws an
# approximate sample from the GP posterior which is a surrogate for the objective
# function. We instantiate our utility function builder below, specifying the number of
# Random Fourier features to use when constructing the approximate samples from the GP posterior:

# %%
utility_function_builder = ThompsonSampling(num_features=500)

# %% [markdown]
# We also need a method for maximising the utility function. Since `ThompsonSampling` is
# classed as a `SinglePointUtilityFunction`, we can use the
# `ContinuousSinglePointUtilityMaximizer` to maximise it. This requires the user to
# specify `num_initial_samples` and `num_restarts` when instantiating it. This first
# queries the utility function at `num_initial_samples` points, and then uses the best of
# these points as a starting point for L-BFGS-B, a gradient-based optimiser, to further
# refine. This is repeated `num_restarts` times, each time sampling a different initial set
# of `num_initial_samples` and the best point found is returned. We'll instantiate our
# maximiser below:

# %%
acquisition_maximizer = ContinuousSinglePointUtilityMaximizer(
    num_initial_samples=100, num_restarts=1
)

# %% [markdown]
#
# It is worth noting that `ThompsonSampling` is not the only utility function we could use,
# since our module also provides e.g. `ProbabilityOfImprovement`, `ExpectedImprovment`,
# which were briefly discussed in [our previous introduction to Bayesian optimisation](https://docs.jaxgaussianprocesses.com/examples/bayesian_optimisation/).


# %% [markdown]
# ## Putting it All Together with the Decision Maker
#
# We now have all the ingredients ready for our Bayesian optimisation loop, so let's put
# all the logic together using the `UtilityDrivenDecisionMaker` class provided by the
# `decision_making` module. This class has 3 core methods:
# 1. `ask` - This method is used to decide which point(s) to query next.
# 2. `tell` - This method is used to tell the decision maker the results from querying the
#    black-box function at the points returned by `ask`, and will often update GP
#    posteriors in light of this data.
# 3. `run` - This is used to run the decision making loop for a specified number of
#    iterations, alternating between `ask` and `tell`.
#
# For many decision making problems, the logic provided in the
# `UtilityDrivenDecisionMaker` will be sufficient, and is a convenient way of gluing the
# various bits of machinery involved in sequential decision making together. However, for
# more exotic decision making loops, it is easy for the user to define their own decision
# maker class by inheriting from the `AbstractDecisionMaker` class and defining their own
# `ask`, `tell` and `run` methods.
#
# However, we do also provide the user with some additional flexibility when using the
# `UtilityDrivenDecisionMaker` class. Often we may wish to perform certain actions after
# the `ask` step and the `tell` step, such as plotting the acquisition function and the
# point chosen to be queried for debugging purposes. We can do this by passing a list of
# functions to be called at each of these points as the `post_ask` and `post_tell`
# attributes of the `UtilityDrivenDecisionMaker`. Both sets of functions are called with
# the `UtilityDrivenDecisionMaker` as an argument, and so have access to all the
# attributes of the decision maker. The `post_ask` functions are additionally passed the
# most recently queried points too. We'll use this functionality to plot the acquisition
# function and the point chosen to be queried at each step of the decision making loop:


# %%
def plot_bo_iteration(
    dm: UtilityDrivenDecisionMaker, last_queried_points: Float[Array, "B D"]
):
    posterior = dm.posteriors[OBJECTIVE]
    dataset = dm.datasets[OBJECTIVE]
    plt_x = jnp.linspace(0, 1, 1000).reshape(-1, 1)
    forrester_y = forrester(plt_x.squeeze(axis=-1))
    utility_fn = dm.current_utility_functions[0]
    sample_y = -utility_fn(plt_x)

    latent_dist = posterior.predict(plt_x, train_data=dataset)
    predictive_dist = posterior.likelihood(latent_dist)

    predictive_mean = predictive_dist.mean()
    predictive_std = predictive_dist.stddev()

    fig, ax = plt.subplots()
    ax.plot(plt_x.squeeze(), predictive_mean, label="Predictive Mean", color=cols[1])
    ax.fill_between(
        plt_x.squeeze(),
        predictive_mean - 2 * predictive_std,
        predictive_mean + 2 * predictive_std,
        alpha=0.2,
        label="Two sigma",
        color=cols[1],
    )
    ax.plot(
        plt_x.squeeze(),
        predictive_mean - 2 * predictive_std,
        linestyle="--",
        linewidth=1,
        color=cols[1],
    )
    ax.plot(
        plt_x.squeeze(),
        predictive_mean + 2 * predictive_std,
        linestyle="--",
        linewidth=1,
        color=cols[1],
    )
    ax.plot(plt_x.squeeze(), sample_y, label="Posterior Sample")
    ax.plot(
        plt_x.squeeze(),
        forrester_y,
        label="Forrester Function",
        color=cols[0],
        linestyle="--",
        linewidth=2,
    )
    ax.axvline(x=0.757, linestyle=":", color=cols[3], label="True Optimum")
    ax.scatter(dataset.X, dataset.y, label="Observations", color=cols[2], zorder=2)
    ax.scatter(
        last_queried_points[0],
        -utility_fn(last_queried_points[0][None, ...]),
        label="Posterior Sample Optimum",
        marker="*",
        color=cols[3],
        zorder=3,
    )
    ax.legend(loc="center left", bbox_to_anchor=(0.950, 0.5))
    plt.show()


# %% [markdown]
# Now let's put it all together and run our decision making loop for 6 iterations, with a
# batch size of 1:

# %%
dm = UtilityDrivenDecisionMaker(
    search_space=search_space,
    posterior_handlers=posterior_handlers,
    datasets=initial_datasets,
    utility_function_builder=utility_function_builder,
    utility_maximizer=acquisition_maximizer,
    batch_size=1,
    key=key,
    post_ask=[plot_bo_iteration],
    post_tell=[],
)

results = dm.run(
    6,
    black_box_function_evaluator=function_evaluator,
)

# %% [markdown]
# We can see that our `DecisionMaker` is successfully able to find the minimimizer of the
# black box function!
#
#

# %% [markdown]
# ## Conclusions

# %% [markdown]
# In this notebook we have provided an introduction to the new `decision_making` module of
# GPJax. We have demonstrated how one may use the abstractions provided by this module to
# implement a Bayesian optimisation loop, and have also highlighted some of the
# flexibility provided by the module. We hope that this module will provide a useful
# framework for solving a wide range of sequential decision making problems, and that it
# will be easy for users to extend the functionality provided by the module to suit their
# needs!
#
# We should note that the `decision_making` module is still in its early stages, and so
# whilst we hope to avoid making breaking changes to it, they may occur as the module
# evolves and more advanced functionality is implemented. If people have any feedback or
# features they would like to implement/see implemented, feel free to open an issue on the
# [GPJax GitHub page](https://github.com/JaxGaussianProcesses/GPJax/issues).
#

# %% [markdown]
# ## System Configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Christie'
